@@ -1,21 +1,18 @@
-set -euf
+#!/bin/bash
+
+set -euxo pipefail
 
 # This is a fix for user.Current issue
 export USER="${USER:-conda}"
 export HOME="${HOME:-$(cd $SRC_DIR/..;pwd)}"
 
 
-#
 # Use precompiled bootstrap
-case $ARCH in
-  aarch64|ppc64le)
-    export GOROOT_BOOTSTRAP=$SRC_DIR/go-bootstrap
-    ;;
-  *)
-    export GOCACHE=off
-    ;;
-esac
-
+if [[ ${target_platform} != "linux-64" ]]; then
+  export GOROOT_BOOTSTRAP=$SRC_DIR/go-bootstrap
+else
+  export GOCACHE=off
+fi
 
 # Do not use GOROOT_FINAL. Otherwise, every conda environment would
 # need its own non-hardlinked copy of the go (+100MB per env).
@@ -25,6 +22,14 @@ esac
 export GOROOT=$SRC_DIR/go
 
 
+if [[ "${target_platform}" == "osx-64" ]]; then
+  export GOOS=darwin
+  export GOARCH=amd64
+elif [[ "${target_platform}" == "osx-arm64" ]]; then
+  export GOOS=darwin
+  export GOARCH=arm64
+fi
+
 # Print diagnostics before building
 env | sort
 
@@ -52,4 +57,9 @@ rm -rf "${PREFIX}"/go/test/fixedbugs/issue27836.dir
 # Right now, it's just go and gofmt, but might be more in the future!
 # We don't move files, and instead rely on soft-links
 mkdir -p ${PREFIX}/bin && pushd $_
-find ../go/bin -type f -exec ln -s {} . \;
+
+if [[ "${build_platform}" != "${target_platform}" ]]; then
+  find ../go/bin/${GOOS}_${GOARCH} -type f -exec ln -s {} . \;
+else
+  find ../go/bin -type f -exec ln -s {} . \;
+fi
